// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

import 'room/room.dart';

/// TODO: Sort order: Alphabetical (a-z) on [User.state.name], otherwise
/// [User.id].
class Members extends Storables<UserId, User> {
  final _memory = List<UserState>();
  final Room _room;

  /// Amount of members.
  ///
  /// Note that this does not reflect how much members are
  /// in memory or in the store. This is the actual count of
  /// members in the [Room].
  ///
  /// Is -1 if unknown.
  int count = -1;

  Members(this._room, this.count, String scope) : super(scope);

  @override
  @protected
  Iterable<User> getFromMemory() {
    return Set.of(_memory.map((state) => User.fromState(state).id))
        .map((id) => getStateOf(id))
        .whereType<User>();
  }

  @override
  @protected
  void setInMemory(Iterable<User> users) =>
      _memory.addAll(users.map((u) => u.state));

  @override
  @protected
  void setInMemoryIfAbsent(User user, User Function() ifAbsent) {
    final added = _memory.contains(user);

    _memory.add(user.state);
    if (added) {
      ifAbsent();
    }
  }

  @override
  @protected
  void removeInMemory(UserId id) =>
      _memory.removeWhere((state) => state.isStateOf(id));

  @override
  @protected
  Future<User> getSingleFromStore(UserId id) => store?.getMember(_room.id, id);

  @override
  @protected
  Future<void> setInStore(Iterable<User> users) =>
      store?.addMembers(_room.id, users);

  @override
  Future<Iterable<User>> getFromStore({User from, int count}) async {
    return store?.getMembers(_room.id, from: from, count: count);
  }

  @override
  Future<int> getSavedInStoreCount() => store?.getMemberCount(_room.id);

  @override
  Future<void> removeInStore(UserId id) => null;

  @override
  Future<Iterable<User>> getFromRemote({int count}) async =>
      _room.retrieveRemoteMembers().then((members) => members.take(count));

  @override
  Iterable<User> sort(Iterable<User> items) {
    final list = items.toList(growable: false);

    return list
      ..sort((a, b) {
        final aName = a.name ?? a.id.toString();
        final bName = b.name ?? b.id.toString();

        return aName.compareTo(bName);
      });
  }

  @override
  bool isMax(int count) => count >= this.count;

  @override
  FutureOr<User> operator [](UserId id) => getStateOf(id);

  /// Returns a [User] with the given [id]. If [at] is provided,
  /// returns a [User] with a `state` that was set at that time.
  ///
  /// If [at] is `null` (default), the most recent state is returned.
  FutureOr<User> getStateOf(UserId id, {DateTime at}) {
    if (at == null) {
      at = DateTime.now();
    }

    final states =
        _memory.where((state) => state.isStateOf(id)).toList(growable: false);

    // If even the first user state is before `at`,
    // return that one.
    if (states.isNotEmpty) {
      final state = states.first;

      if (at.millisecondsSinceEpoch < state.since.millisecondsSinceEpoch) {
        return User.fromState(state);
      }
    }

    for (var i = 0; i < states.length; i++) {
      final current = states[i];
      final next = i + 1 < states.length ? states[i + 1] : null;

      final currentIsBefore =
          current.since.millisecondsSinceEpoch <= at.millisecondsSinceEpoch;

      final nextIsAfterOrNull = next == null ||
          next.since.millisecondsSinceEpoch > at.millisecondsSinceEpoch;

      if (currentIsBefore && nextIsAfterOrNull) {
        return User.fromState(current);
      }
    }

    // If not in memory, return from store
    return store?.getMember(_room.id, id, at: at);
  }
}
