// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:http/http.dart';
import 'package:http/testing.dart';
import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:test/test.dart';
import 'package:url/url.dart';

void testHomeserver() {
  group('Homeserver', () {
    group('.login ', () {
      test('returns LocalUser if successful', () async {
        final hs = Homeserver(
          Url.parse('https://pattle.im'),
          httpClient: MockClient((request) async {
            return Response(
              json.encode({
                'user_id': '@pat:pattle.im',
                'access_token': 'abc123',
                'device_id': 'GHTYAJCE',
                'well_known': {
                  'm.homeserver': {
                    'base_url': 'https://pattle.im',
                  },
                  'm.identity_server': {
                    'base_url': 'https://id.pattle.im',
                  },
                },
              }),
              200,
            );
          }),
        );

        final user = await hs.login(
          Username('pat'),
          'password',
          store: MemoryStore(),
        );

        expect(user, isNotNull);
        expect(user.id.toString(), matches('@pat:pattle.im'));
        expect(user.accessToken, equals('abc123'));
      });

      test('returns LocalUser with given device id if none given', () async {
        final hs = Homeserver(
          Url.parse('https://pattle.im'),
          httpClient: MockClient((request) async {
            return Response(
              json.encode({
                'user_id': '@pat:pattle.im',
                'access_token': 'abc123',
                'device_id': 'GHTYAJCE',
                'well_known': {
                  'm.homeserver': {
                    'base_url': 'https://pattle.im',
                  },
                  'm.identity_server': {
                    'base_url': 'https://id.pattle.im',
                  },
                },
              }),
              200,
            );
          }),
        );

        final user = await hs.login(
          Username('pat'),
          'password',
          store: MemoryStore(),
        );

        expect(user.currentDevice.id.toString(), equals('GHTYAJCE'));
      });

      test('returns LocalUser with set device id by caller', () async {
        final hs = Homeserver(
          Url.parse('https://pattle.im'),
          httpClient: MockClient((request) async {
            return Response(
              json.encode({
                'user_id': '@pat:pattle.im',
                'access_token': 'abc123',
                'device_id': 'GHTYAJCE',
                'well_known': {
                  'm.homeserver': {
                    'base_url': 'https://pattle.im',
                  },
                  'm.identity_server': {
                    'base_url': 'https://id.pattle.im',
                  },
                },
              }),
              200,
            );
          }),
        );

        final user = await hs.login(
          Username('pat'),
          'password',
          store: MemoryStore(),
          device: Device(
            id: DeviceId('BLABLA'),
            name: 'Pattle Android',
          ),
        );

        expect(user.currentDevice.id.toString(), equals('BLABLA'));
        expect(user.currentDevice.name, equals('Pattle Android'));
      });

      test('throws ForbiddenException if password is wrong', () async {
        final hs = Homeserver(
          Url.parse('https://pattle.im'),
          httpClient: MockClient((request) async {
            return Response(
              json.encode({'errcode': 'M_FORBIDDEN'}),
              403,
            );
          }),
        );

        expect(
          hs.login(
            Username('pat'),
            'password',
            store: MemoryStore(),
          ),
          throwsA(isA<ForbiddenException>()),
        );
      });
    });

    group('.fromWellKnown ', () {
      test(' will throw exception if .well-known base_url is null', () async {
        expect(
          Homeserver.fromWellKnown(
            Url.parse('https://pattle.im'),
            httpClient: MockClient((request) async {
              return Response(
                json.encode({
                  'm.homeserver': {'base_url': null},
                }),
                200,
              );
            }),
          ),
          throwsA(isA<WellKnownFailPromptException>()),
        );
      });

      test(' will remove slash suffix from url', () async {
        final hs = await Homeserver.fromWellKnown(
          Url.parse('https://pattle.im'),
          httpClient: MockClient((request) async {
            return Response(
              json.encode({
                'm.homeserver': {
                  'base_url': 'https://pattle.im/',
                },
              }),
              200,
            );
          }),
        );

        expect(hs.url.toString(), isNot(endsWith('/')));
      });
    });
  });

  // TODO: Handle 400
}
